import copy

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils import timezone
from django.utils.formats import localize
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, can_proceed, transition

from . import constants, versionables
from .conditions import (
    Conditions,
    draft_is_locked,
    draft_is_not_locked,
    in_state,
    is_not_locked,
    user_can_change,
    user_can_publish,
    user_can_unlock,
)
from .conf import ALLOW_DELETING_VERSIONS, LOCK_VERSIONS
from .operations import send_post_version_operation, send_pre_version_operation

try:
    from djangocms_internalsearch.helpers import emit_content_change
except ImportError:
    emit_content_change = None


not_draft_error = _("Version is not a draft")
lock_error_message = _("Action Denied. The latest version is locked by {user}")
lock_draft_error_message = _("Action Denied. The draft version is locked by {user}")
permission_error_message = _("You do not have permission to perform this action")

def allow_deleting_versions(collector, field, sub_objs, using):
    if ALLOW_DELETING_VERSIONS:
        models.SET_NULL(collector, field, sub_objs, using)
    else:
        models.PROTECT(collector, field, sub_objs, using)


class VersionQuerySet(models.QuerySet):
    def get_for_content(self, content_object):
        """Returns Version object corresponding to provided content object
        """
        if hasattr(content_object, "_version_cache"):
            return content_object._version_cache
        versionable = versionables.for_content(content_object)
        version = self.get(
            object_id=content_object.pk, content_type__in=versionable.content_types
        )
        content_object._version_cache = version
        return version

    def filter_by_grouper(self, grouper_object):
        """Returns a list of Version objects for the provided grouper
        object
        """
        versionable = versionables.for_grouper(grouper_object)
        return self.filter_by_grouping_values(
            versionable, **{versionable.grouper_field_name: grouper_object}
        )

    def filter_by_grouping_values(self, versionable, **kwargs):
        """Returns a list of Version objects for the provided grouping
        values (unique grouper version list)
        """
        content_objects = versionable.for_grouping_values(**kwargs)
        return self.filter(
            object_id__in=content_objects, content_type__in=versionable.content_types
        )

    def filter_by_content_grouping_values(self, content):
        """Returns a list of Version objects for grouping values taken
        from provided content object. In other words:
        it uses the content instance property values as filter parameters
        """
        versionable = versionables.for_content(content)
        content_objects = versionable.for_content_grouping_values(content)
        return self.filter(
            object_id__in=content_objects, content_type__in=versionable.content_types
        )


class Version(models.Model):

    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    modified = models.DateTimeField(default=timezone.now, verbose_name=_("Modified"))
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, verbose_name=_("author")
    )
    number = models.CharField(max_length=11, verbose_name="#")
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.PROTECT,
        related_name="cms_versions"
    )
    object_id = models.PositiveIntegerField()
    content = GenericForeignKey("content_type", "object_id")
    state = FSMField(
        default=constants.DRAFT,
        choices=constants.VERSION_STATES,
        verbose_name=_("status"),
        protected=True,
    )
    locked_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,  # Deleting a user removes the lock
        null=True,
        default=None,
        verbose_name=_("locked by"),
        related_name="locking_users",
    )

    source = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=allow_deleting_versions,
        verbose_name=_("source"),
    )
    objects = VersionQuerySet.as_manager()

    class Meta:
        unique_together = ("content_type", "object_id")
        permissions = (
            ("delete_versionlock", "Can unlock verision"),
        )

    def __str__(self):
        return f"Version #{self.pk}"

    def verbose_name(self):
        return _("Version #{number} ({state} {date})").format(
            number=self.number,
            state=dict(constants.VERSION_STATES)[self.state],
            date=localize(self.created, settings.DATETIME_FORMAT),
        )

    def short_name(self):
        return _("Version #{number} ({state})").format(
            number=self.number, state=dict(constants.VERSION_STATES)[self.state]
        )

    def locked_message(self):
        if self.locked_by:
            return _("Locked by %(user)s") % {"user": self.locked_by}
        return ""

    def delete(self, using=None, keep_parents=False):
        """Deleting a version deletes the grouper
        as well if we are deleting the last version."""

        def get_grouper_name(ContentModel, GrouperModel):
            for field in ContentModel._meta.fields:
                if getattr(field, "related_model", None) == GrouperModel:
                    return field.name

        grouper = self.grouper
        ContentModel = self.content._meta.model

        grouper_name = get_grouper_name(ContentModel, grouper._meta.model)
        querydict = {f"{grouper_name}__pk": grouper.pk}
        count = ContentModel._original_manager.filter(**querydict).count()

        self.content.delete()
        deleted = super().delete(using=using, keep_parents=keep_parents)
        deleted[1]["last"] = False
        if count == 1:
            grouper.delete()
            deleted[1]["last"] = True
        return deleted

    def save(self, **kwargs):
        created = not self.pk
        # On version creation
        if created:
            # trigger pre operation signal
            action_token = send_pre_version_operation(
                constants.OPERATION_DRAFT, version=self
            )
            # Set the version number
            self.number = self.make_version_number()
        if self.pk is None and self.state == constants.DRAFT:
            # A new draft version is locked by default
            if LOCK_VERSIONS and self.locked_by is None:
                # create a lock
                self.locked_by = self.created_by
        elif self.state != constants.DRAFT:
            # A any other state than draft has no lock, an existing lock should be removed
            self.locked_by = None

        super().save(**kwargs)
        # Only one draft version is allowed per unique grouping values.
        # Set all other drafts to archived
        if self.state == constants.DRAFT:
            if created:
                pks_for_grouping_values = self.versionable.for_content_grouping_values(
                    self.content
                ).values_list("pk", flat=True)
                to_archive = Version.objects.exclude(pk=self.pk).filter(
                    state=constants.DRAFT,
                    object_id__in=pks_for_grouping_values,
                    content_type=self.content_type,
                )
                for version in to_archive:
                    version.archive(self.created_by)
                on_draft_create = self.versionable.on_draft_create
                if on_draft_create:
                    on_draft_create(self)
                # trigger post operation signal
                send_post_version_operation(
                    constants.OPERATION_DRAFT, version=self, token=action_token
                )
            if emit_content_change:
                emit_content_change(self.content, created=created)

    def make_version_number(self):
        """
        Create a version number for each version
        """
        # Get the latest version object
        latest_version = (
            Version.objects.filter_by_content_grouping_values(self.content)
            .order_by("-pk")
            .first()
        )
        # If no previous version exists start at 1
        if not latest_version:
            return 1
        return int(latest_version.number) + 1

    @property
    def versionable(self):
        """Helper property to get the versionable for the content type
        of the version
        """
        return versionables.for_content(self.content)

    def convert_to_proxy(self):
        """Returns a copy of current Version object, but as an instance
        of its correct proxy model"""

        new_obj = copy.deepcopy(self)
        new_obj.__class__ = self.versionable.version_model_proxy
        return new_obj

    @property
    def grouper(self):
        """Helper property to get the grouper for the version
        """
        return getattr(self.content, self.versionable.grouper_field_name)

    @transaction.atomic
    def copy(self, created_by):
        """Creates a new Version object, with a copy of the related
        content object.
        Allows customization of how the content object will be copied
        when specified in cms_config.py

        This method needs to be run in a transaction due to the fact that if
        models are partially created in the copy method a version is not attached.
        It needs to be that if anything goes wrong we should roll back the entire task.
        We shouldn't leave this to package developers to know to add this feature
        because not doing so leaves the db and versioning in a corrupt state where
        content models exist without a version.
        """
        copy_function = versionables.for_content(self.content).copy_function
        new_content = copy_function(self.content)

        new_version = Version.objects.create(
            content=new_content, source=self, created_by=created_by,
            **({"locked_by": created_by} if LOCK_VERSIONS else {}),
        )
        return new_version

    check_archive = Conditions(
        [
            user_can_change(permission_error_message),
            in_state([constants.DRAFT], _("Version is not in draft state")),
            is_not_locked(lock_error_message),
        ]
    )

    def can_be_archived(self):
        return can_proceed(self._set_archive)

    def archive(self, user):
        """Change state to ARCHIVED"""
        # trigger pre operation signal
        action_token = send_pre_version_operation(
            constants.OPERATION_ARCHIVE, version=self
        )
        self._set_archive(user)
        self.modified = timezone.now()
        self.save()
        StateTracking.objects.create(
            version=self,
            old_state=constants.DRAFT,
            new_state=constants.ARCHIVED,
            user=user,
        )
        on_archive = self.versionable.on_archive
        if on_archive:
            on_archive(self)
        # trigger post operation signal
        send_post_version_operation(
            constants.OPERATION_ARCHIVE, version=self, token=action_token
        )
        if emit_content_change:
            emit_content_change(self.content)

    @transition(
        field=state,
        source=constants.DRAFT,
        target=constants.ARCHIVED,
        permission=check_archive.as_bool,
    )
    def _set_archive(self, user):
        """State machine transition method for moving version
        from DRAFT to ARCHIVED state.

        Please refrain from modifying data in this method, as
        state change is not guaranteed to be saved (making it
        possible to be left with inconsistent data)"""
        pass

    check_publish = Conditions(
        [
            user_can_publish(permission_error_message),
            in_state([constants.DRAFT], _("Version is not in draft state")),
        ]
    )

    def can_be_published(self):
        return can_proceed(self._set_publish)

    def publish(self, user):
        """Change state to PUBLISHED and unpublish currently
        published versions"""
        # trigger pre operation signal
        action_token = send_pre_version_operation(
            constants.OPERATION_PUBLISH, version=self
        )
        self._set_publish(user)
        self.modified = timezone.now()
        self.save()
        StateTracking.objects.create(
            version=self,
            old_state=constants.DRAFT,
            new_state=constants.PUBLISHED,
            user=user,
        )
        # Only one published version is allowed per unique grouping values.
        # Set all other published versions to unpublished
        pks_for_grouping_values = self.versionable.for_content_grouping_values(
            self.content
        ).values_list("pk", flat=True)
        to_unpublish = Version.objects.exclude(pk=self.pk).filter(
            state=constants.PUBLISHED,
            object_id__in=pks_for_grouping_values,
            content_type=self.content_type,
        )
        for version in to_unpublish:
            version.unpublish(user, to_be_published=self)
        on_publish = self.versionable.on_publish
        if on_publish:
            on_publish(self)
        # trigger post operation signal
        send_post_version_operation(
            constants.OPERATION_PUBLISH,
            version=self,
            token=action_token,
            unpublished=list(to_unpublish),
        )
        if emit_content_change:
            emit_content_change(self.content)

    @transition(
        field=state,
        source=constants.DRAFT,
        target=constants.PUBLISHED,
        permission=check_publish.as_bool,
    )
    def _set_publish(self, user):
        """State machine transition method for moving version
        from DRAFT to PUBLISHED state.

        Please refrain from modifying data in this method, as
        state change is not guaranteed to be saved (making it
        possible to be left with inconsistent data)"""
        pass

    check_unpublish = Conditions([
        user_can_publish(permission_error_message),
        in_state([constants.PUBLISHED], _("Version is not in published state")),
        draft_is_not_locked(lock_draft_error_message),
    ])

    def can_be_unpublished(self):
        return can_proceed(self._set_unpublish)

    def unpublish(self, user, to_be_published=None):
        """Change state to UNPUBLISHED"""
        # trigger pre operation signal
        action_token = send_pre_version_operation(
            constants.OPERATION_UNPUBLISH, version=self, to_be_published=to_be_published
        )
        self._set_unpublish(user)
        self.modified = timezone.now()
        self.save()
        StateTracking.objects.create(
            version=self,
            old_state=constants.PUBLISHED,
            new_state=constants.UNPUBLISHED,
            user=user,
        )
        on_unpublish = self.versionable.on_unpublish
        if on_unpublish:
            on_unpublish(self)
        # trigger post operation signal
        send_post_version_operation(
            constants.OPERATION_UNPUBLISH,
            version=self,
            token=action_token,
            to_be_published=to_be_published,
        )
        if emit_content_change:
            emit_content_change(self.content)

    @transition(
        field=state,
        source=constants.PUBLISHED,
        target=constants.UNPUBLISHED,
        permission=check_unpublish.as_bool,
    )
    def _set_unpublish(self, user):
        """State machine transition method for moving version
        from PUBLISHED to UNPUBLISHED state.

        Please refrain from modifying data in this method, as
        state change is not guaranteed to be saved (making it
        possible to be left with inconsistent data)"""
        pass

    def has_publish_permission(self, user) -> bool:
        """
        Check if the given user has permission to publish.

        Args:
            user (User): The user to check for permission.

        Returns:
            bool: True if the user has publish permission, False otherwise.
        """
        return self._has_permission("publish", user)

    def has_change_permission(self, user) -> bool:
        """
        Check whether the given user has permission to change the object.

        Parameters:
            user (User): The user for which permission needs to be checked.

        Returns:
            bool: True if the user has permission to change the object, False otherwise.
        """
        return self._has_permission("change", user)

    def _has_permission(self, perm: str, user) -> bool:
        """
        Check if the user has the specified permission for the content by
        checking the content's has_publish_permission, has_placeholder_change_permission,
        or has_change_permission methods.

        Falls back to Djangos change permission for the content object.
        """
        if perm == "publish" and hasattr(self.content, "has_publish_permission"):
            # First try explicit publish permission
            return self.content.has_publish_permission(user)
        if hasattr(self.content, "has_change_permission"):
            # First fallback: change permissions
            return self.content.has_change_permission(user)
        if hasattr(self.content, "has_placeholder_change_permission"):
            # Second fallback: placeholder change permissions - works for PageContent
            return self.content.has_placeholder_change_permission(user)
        # final fallback: Django perms
        return user.has_perm(f"{self.content_type.app_label}.change_{self.content_type.model}")

    check_modify = Conditions(
        [
            in_state([constants.DRAFT], not_draft_error),
            draft_is_not_locked(lock_draft_error_message),
            user_can_unlock(permission_error_message),
        ]
    )
    check_revert = Conditions(
        [
            user_can_change(permission_error_message),
            in_state(
                [constants.ARCHIVED, constants.UNPUBLISHED],
                _("Version is not in archived or unpublished state"),
            ),
            draft_is_not_locked(lock_draft_error_message),
        ]
    )
    check_discard = Conditions(
        [
            in_state([constants.DRAFT], not_draft_error),
            is_not_locked(lock_error_message),
        ]
    )
    check_edit_redirect = Conditions(
        [
            in_state(
                [constants.DRAFT, constants.PUBLISHED],
                _("Version is not in draft or published state"),
            ),
            draft_is_not_locked(lock_draft_error_message),
        ]
    )
    check_lock = Conditions(
        [
            in_state([constants.DRAFT], not_draft_error),
            is_not_locked(_("Version is already locked"))
        ]
    )
    check_unlock = Conditions(
        [
            in_state([constants.DRAFT, constants.PUBLISHED], not_draft_error),
            draft_is_locked(_("Draft version is not locked"))

        ]
    )


class StateTracking(models.Model):
    version = models.ForeignKey(Version, on_delete=models.CASCADE)
    date = models.DateTimeField(auto_now_add=True)
    old_state = models.CharField(max_length=100, choices=constants.VERSION_STATES)
    new_state = models.CharField(max_length=100, choices=constants.VERSION_STATES)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)



from __future__ import annotations

from functools import lru_cache

from django import forms
from django.contrib.admin.widgets import AutocompleteSelect

from . import versionables


class VersionAutocompleteSelect(AutocompleteSelect):
    def optgroups(self, name: str, value: str, attr: dict | None = None):
        default = (None, [], 0)
        default[1].append(self.create_option(name, "", "", False, 0))
        return [default]


class VersionContentChoiceField(forms.ModelChoiceField):
    """Form field used to display a list of grouper instances"""

    def __init__(self, *args, model=None, admin_site=None, **kwargs):
        self.language = kwargs.pop("language")
        self.predefined_label_method = kwargs.pop("option_label_override")
        if getattr(admin_site._registry.get(model), "search_fields", []):
            # If the model is registered in the admin, use the autocomplete widget
            kwargs.setdefault("widget", VersionAutocompleteSelect(
                model._meta.get_field(versionables.for_content(model).grouper_field_name),
                admin_site=admin_site,
            ))
        super().__init__(*args, **kwargs)

    def label_from_instance(self, obj):
        """Overridden to allow customizing the labels of the groupers"""
        if self.predefined_label_method:
            return self.predefined_label_method(obj, self.language)
        else:
            return super().label_from_instance(obj)


@lru_cache
def grouper_form_factory(content_model, language=None, admin_site=None):
    """Returns a form class used for selecting a grouper to see versions of.
    Form has a single field - grouper - which is a model choice field
    with available grouper objects for specified content model.

    :param content_model: Content model class
    :param language: Language
    """
    if admin_site is None:
        from django.contrib.admin import site
        admin_site = site

    versionable = versionables.for_content(content_model)
    return type(
        content_model.__name__ + "GrouperForm",
        (forms.Form,),
        {
            "_content_model": content_model,
            versionable.grouper_field_name: VersionContentChoiceField(
                label=versionable.grouper_model._meta.verbose_name.capitalize(),
                queryset=versionable.grouper_model.objects.all(),
                option_label_override=versionable.grouper_selector_option_label,
                admin_site=admin_site,
                model=content_model,
                language=language,
            ),
        },
    )



from cms import constants as cms_constants
from cms.apphook_pool import apphook_pool
from cms.cms_menus import CMSMenu as OriginalCMSMenu, get_visible_nodes
from cms.models import Page

try:
    from cms.models import TreeNode
except ImportError:
    TreeNode = None
from cms.toolbar.utils import get_object_preview_url, get_toolbar_from_request
from cms.utils.page import get_page_queryset
from django.apps import apps
from menus.base import Menu, NavigationNode
from menus.menu_pool import menu_pool

from . import conf, constants


class CMSVersionedNavigationNode(NavigationNode):
    def is_selected(self, request):
        try:
            page_id = request.current_page.pk
        except AttributeError:
            return False
        return page_id == self.id


def _get_attrs_for_node(renderer, page_content):
    page = page_content.page
    language = renderer.request_language
    attr = {
        "is_page": True,
        "soft_root": page_content.soft_root,
        "auth_required": page.login_required,
        "reverse_id": page.reverse_id,
    }
    limit_visibility_in_menu = page_content.limit_visibility_in_menu

    if limit_visibility_in_menu is cms_constants.VISIBILITY_ALL:
        attr["visible_for_authenticated"] = True
        attr["visible_for_anonymous"] = True
    else:
        attr["visible_for_authenticated"] = (
            limit_visibility_in_menu == cms_constants.VISIBILITY_USERS
        )
        attr["visible_for_anonymous"] = (
            limit_visibility_in_menu == cms_constants.VISIBILITY_ANONYMOUS
        )

    attr["is_home"] = page.is_home
    extenders = []

    if page.navigation_extenders:
        if page.navigation_extenders in renderer.menus:
            extenders.append(page.navigation_extenders)
        elif f"{page.navigation_extenders}:{page.pk}" in renderer.menus:
            extenders.append(f"{page.navigation_extenders}:{page.pk}")

    if page.application_urls:
        app = apphook_pool.get_apphook(page.application_urls)

        if app:
            extenders += app.get_menus(page, language)

    exts = []

    for ext in extenders:
        if hasattr(ext, "get_instances"):
            exts.append(f"{ext.__name__}:{page.pk}")
        elif hasattr(ext, "__name__"):
            exts.append(ext.__name__)
        else:
            exts.append(ext)

    if exts:
        attr["navigation_extenders"] = exts

    attr["redirect_url"] = page_content.redirect

    return attr


class CMSMenu(Menu):
    """This is a legacy class used by django CMS 4.0 and django CMS 4.1.0 only. Its language
    fallback mechanism does not comply with django CMS' core's. Also, it is by far slower
    than django CMS core's. As of django CMS 4.1.1, this class is by default deactivated.

    See https://discord.com/channels/800813886689247262/1204047551570120755 for more information."""
    def get_nodes(self, request):
        site = self.renderer.site
        language = self.renderer.request_language
        pages_qs = get_page_queryset(site).select_related("node")
        visible_pages_for_user = get_visible_nodes(request, pages_qs, site)

        if not visible_pages_for_user:
            return []

        cms_extension = apps.get_app_config("djangocms_versioning").cms_extension
        toolbar = get_toolbar_from_request(request)
        edit_or_preview = toolbar.edit_mode_active or toolbar.preview_mode_active
        menu_nodes = []
        node_id_to_page = {}
        homepage_content = None

        # Depending on the toolbar mode, we need to get the correct version.
        # On edit or preview mode: return DRAFT,
        # if DRAFT does not exist then return PUBLISHED.
        # On public mode: return PUBLISHED.
        if edit_or_preview:
            states = [constants.DRAFT, constants.PUBLISHED]
        else:
            states = [constants.PUBLISHED]

        versionable_item = cms_extension.versionables_by_grouper[Page]
        versioned_page_contents = (
            versionable_item.content_model._base_manager.filter(
                language=language, page__in=pages_qs, versions__state__in=states
            )
            .order_by("page__node__path" if TreeNode else "page__path", "versions__state")
            .select_related("page", "page__node" if TreeNode else "page")
            .prefetch_related("versions")
        )
        added_pages = []

        for page_content in versioned_page_contents:
            page = page_content.page

            if page not in visible_pages_for_user:
                # The page is restricted for the user.
                # Therefore, we avoid adding it to the menu.
                continue

            version = page_content.versions.all()[0]

            if (
                page.pk in added_pages
                and edit_or_preview
                and version.state == constants.PUBLISHED
            ):
                # Page content is already added. This is the case where you
                # have both draft and published and in edit/preview mode.
                # We give priority to draft which is already sorted by the query.
                # Therefore we ignore the published version.
                continue

            page_tree_node = page.node
            parent_id = node_id_to_page.get(page_tree_node.parent_id)

            if page_tree_node.parent_id and not parent_id:
                # If the parent page is not available,
                # we skip adding the menu node.
                continue

            # Construct the url based on the toolbar mode.
            if edit_or_preview:
                url = get_object_preview_url(page_content)
            else:
                url = page_content.get_absolute_url()

            # Create the new navigation node.
            new_node = CMSVersionedNavigationNode(
                id=page.pk,
                attr=_get_attrs_for_node(self.renderer, page_content),
                title=page_content.menu_title or page_content.title,
                url=url,
                visible=page_content.in_navigation,
            )

            if not homepage_content:
                # Set the home page content.
                homepage_content = page_content if page.is_home else None

            cut_homepage = homepage_content and not homepage_content.in_navigation

            if cut_homepage and parent_id == homepage_content.page.pk:
                # When the homepage is hidden from navigation,
                # we need to cut all its direct children from it.
                new_node.parent_id = None
            else:
                new_node.parent_id = parent_id

            node_id_to_page[page_tree_node.pk] = page.pk
            menu_nodes.append(new_node)
            added_pages.append(page.pk)
        return menu_nodes


if conf.ENABLE_MENU_REGISTRATION:
    # Remove the core djangoCMS CMSMenu and register the new CMSVersionedMenu.
    menu_pool.menus.pop(OriginalCMSMenu.__name__)
    menu_pool.register_menu(CMSMenu)



# Generated by Django 3.2.13 on 2022-05-20 21:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('djangocms_versioning', '0015_version_modified'),
    ]

    operations = [
        migrations.AlterField(
            model_name='version',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='cms_versions', to='contenttypes.contenttype'),
        ),
    ]



from cms.models import PageContent
from cms.test_utils.testcases import CMSTestCase
from cms.test_utils.util.context_managers import signal_tester
from django.dispatch import receiver

from djangocms_versioning import constants
from djangocms_versioning.signals import (
    post_version_operation,
    pre_version_operation,
)
from djangocms_versioning.test_utils import factories


class TestVersioningSignals(CMSTestCase):
    def setUp(self):
        self.superuser = self.get_superuser()

    def test_publish_signals_fired(self):
        """
        When a version is changed to published the correct signals are fired!
        """
        poll = factories.PollFactory()
        version = factories.PollVersionFactory(
            state=constants.DRAFT, content__poll=poll
        )

        with signal_tester(pre_version_operation, post_version_operation) as env:

            version.publish(self.superuser)

            self.assertEqual(env.call_count, 2)

            pre_call_kwargs = env.calls[0][1]
            post_call_kwargs = env.calls[1][1]

            # pre call
            self.assertTrue(pre_call_kwargs["token"] == post_call_kwargs["token"])
            self.assertEqual(pre_call_kwargs["operation"], constants.OPERATION_PUBLISH)
            self.assertEqual(
                pre_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(pre_call_kwargs["obj"], version)
            # post call
            self.assertTrue("token" in post_call_kwargs)
            self.assertEqual(post_call_kwargs["operation"], constants.OPERATION_PUBLISH)
            self.assertEqual(
                post_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(post_call_kwargs["obj"], version)


    def test_publish_signals_fired_with_to_be_published_and_unpublished(self):
        poll = factories.PollFactory()
        version1 = factories.PollVersionFactory(
            state=constants.DRAFT, content__poll=poll
        )
        version2 = version1.copy(self.superuser)

        # Here, we just expect the signals for version 1
        with signal_tester(pre_version_operation, post_version_operation) as env:
            version1.publish(self.superuser)
            self.assertEqual(env.call_count, 2)

        # Here, we expect the signals for the unpublish of version 1 and the
        # publish of version 2.
        with signal_tester(pre_version_operation, post_version_operation) as env:
            version2.publish(self.superuser)
            self.assertEqual(env.call_count, 4)
            version_1_pre_call_kwargs = env.calls[1][1]
            version_2_post_call_kwargs = env.calls[3][1]

            self.assertEqual(version_1_pre_call_kwargs["to_be_published"], version2)
            self.assertEqual(version_2_post_call_kwargs["unpublished"], [version1])


    def test_unpublish_signals_fired(self):
        """
        When a version is changed to unpublished the correct signals are fired!
        """
        poll = factories.PollFactory()
        version = factories.PollVersionFactory(
            state=constants.PUBLISHED, content__poll=poll
        )

        with signal_tester(pre_version_operation, post_version_operation) as env:

            version.unpublish(self.superuser)

            self.assertEqual(env.call_count, 2)

            pre_call_kwargs = env.calls[0][1]
            post_call_kwargs = env.calls[1][1]

            # pre call
            self.assertTrue(pre_call_kwargs["token"] == post_call_kwargs["token"])
            self.assertEqual(
                pre_call_kwargs["operation"], constants.OPERATION_UNPUBLISH
            )
            self.assertEqual(
                pre_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(pre_call_kwargs["obj"], version)
            # post call
            self.assertTrue("token" in post_call_kwargs)
            self.assertEqual(
                post_call_kwargs["operation"], constants.OPERATION_UNPUBLISH
            )
            self.assertEqual(
                post_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(post_call_kwargs["obj"], version)

    def test_archive_signals_fired(self):
        """
        When a version is changed to archived the correct signals are fired!
        """
        poll = factories.PollFactory()
        version = factories.PollVersionFactory(
            state=constants.DRAFT, content__poll=poll
        )

        with signal_tester(pre_version_operation, post_version_operation) as env:

            version.archive(self.superuser)

            self.assertEqual(env.call_count, 2)

            pre_call_kwargs = env.calls[0][1]
            post_call_kwargs = env.calls[1][1]

            # pre call
            self.assertTrue(pre_call_kwargs["token"] == post_call_kwargs["token"])
            self.assertEqual(pre_call_kwargs["operation"], constants.OPERATION_ARCHIVE)
            self.assertEqual(
                pre_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(pre_call_kwargs["obj"], version)
            # post call
            self.assertTrue("token" in post_call_kwargs)
            self.assertEqual(post_call_kwargs["operation"], constants.OPERATION_ARCHIVE)
            self.assertEqual(
                post_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(post_call_kwargs["obj"], version)

    def test_draft_signals_fired(self):
        """
        When a version is set as draft (created) the correct signals are fired!
        """
        poll = factories.PollFactory()

        with signal_tester(pre_version_operation, post_version_operation) as env:

            version = factories.PollVersionFactory(
                state=constants.DRAFT, content__poll=poll
            )

            self.assertEqual(env.call_count, 2)

            pre_call_kwargs = env.calls[0][1]
            post_call_kwargs = env.calls[1][1]

            # pre call
            self.assertTrue(pre_call_kwargs["token"] == post_call_kwargs["token"])
            self.assertEqual(pre_call_kwargs["operation"], constants.OPERATION_DRAFT)
            self.assertEqual(
                pre_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(pre_call_kwargs["obj"], version)
            # post call
            self.assertTrue("token" in post_call_kwargs)
            self.assertEqual(post_call_kwargs["operation"], constants.OPERATION_DRAFT)
            self.assertEqual(
                post_call_kwargs["sender"], version.content_type.model_class()
            )
            self.assertEqual(post_call_kwargs["obj"], version)

    def test_page_signals_publish_unpublish_example(self):
        """
        The example in the docs provides the following example to the page publish and unpublish signals.
        """
        signal_hits = []

        # Signal example
        @receiver(post_version_operation, sender=PageContent)
        def do_something_on_page_publish_unpublsh(*args, **kwargs):

            if (
                kwargs["operation"] == constants.OPERATION_PUBLISH
                or kwargs["operation"] == constants.OPERATION_UNPUBLISH
            ):
                # Storing the state of the operation and object at this moment to compare the state later
                obj = {}
                obj["state"] = kwargs["obj"].state
                signal_hits.append(obj)

        version_1 = factories.PageVersionFactory(
            state=constants.DRAFT, content__template=""
        )
        version_2 = factories.PageVersionFactory(
            state=constants.DRAFT, content__template=""
        )
        version_1.publish(self.superuser)
        version_1.unpublish(self.superuser)
        version_2.archive(self.superuser)

        # Only the publish and unpublish signals should have had an affect
        self.assertEqual(len(signal_hits), 2)
        self.assertEqual(signal_hits[0].get("state"), constants.PUBLISHED)
        self.assertEqual(signal_hits[1].get("state"), constants.UNPUBLISHED)



from django.apps import AppConfig
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _


class VersioningConfig(AppConfig):
    name = "djangocms_versioning"
    verbose_name = _("django CMS Versioning")
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from cms.models import contentmodels, fields
        from cms.signals import post_obj_operation, post_placeholder_operation

        from .conf import LOCK_VERSIONS
        from .handlers import (
            update_modified_date,
            update_modified_date_for_pagecontent,
            update_modified_date_for_placeholder_source,
        )
        from .helpers import is_content_editable, placeholder_content_is_unlocked_for_user

        # Add check to PlaceholderRelationField
        fields.PlaceholderRelationField.default_checks += [is_content_editable]
        if LOCK_VERSIONS:
            fields.PlaceholderRelationField.default_checks += [placeholder_content_is_unlocked_for_user]

        # Remove uniqueness constraint from PageContent model to allow for different versions
        pagecontent_unique_together = tuple(
            set(contentmodels.PageContent._meta.unique_together) - {("language", "page")}
        )
        contentmodels.PageContent._meta.unique_together = pagecontent_unique_together

        # Connect signals
        post_save.connect(update_modified_date, dispatch_uid="versioning")
        post_placeholder_operation.connect(
            update_modified_date_for_placeholder_source, dispatch_uid="versioning"
        )
        post_obj_operation.connect(
            update_modified_date_for_pagecontent, dispatch_uid="versioning"
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-07-30 10:35
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("djangocms_versioning", "0003_version"),
    ]

    operations = [
        migrations.CreateModel(
            name="StateTracking",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                (
                    "old_state",
                    models.CharField(
                        choices=[
                            ("draft", "Draft"),
                            ("published", "Published"),
                            ("unpublished", "Unpublished"),
                            ("archived", "Archived"),
                        ],
                        max_length=100,
                    ),
                ),
                (
                    "new_state",
                    models.CharField(
                        choices=[
                            ("draft", "Draft"),
                            ("published", "Published"),
                            ("unpublished", "Unpublished"),
                            ("archived", "Archived"),
                        ],
                        max_length=100,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="version",
            name="state",
            field=django_fsm.FSMField(
                choices=[
                    ("draft", "Draft"),
                    ("published", "Published"),
                    ("unpublished", "Unpublished"),
                    ("archived", "Archived"),
                ],
                default="draft",
                max_length=50,
                protected=True,
                verbose_name="status",
            ),
        ),
        migrations.AddField(
            model_name="statetracking",
            name="version",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="djangocms_versioning.Version",
            ),
        ),
    ]



from django import template
from django.urls import reverse

from .. import constants, versionables
from ..helpers import version_list_url

register = template.Library()


@register.filter
def url_version_list(content):
    return version_list_url(content)

@register.filter
def url_publish_version(content, user):
    if hasattr(content, "prefetched_versions"):
        version = content.prefetched_versions[0]
    else:
        version = content.versions.first()
    if version:
        if version.check_publish.as_bool(user) and version.can_be_published():
            proxy_model = versionables.for_content(content).version_model_proxy
            return reverse(
                f"admin:{proxy_model._meta.app_label}_{proxy_model.__name__.lower()}_publish",
                args=(version.pk,),
            )
    return ""

@register.filter
def url_new_draft(content, user):
    if hasattr(content, "prefetched_versions"):
        version = content.prefetched_versions[0]
    else:
        version = content.versions.first()
    if version:
        if version.state == constants.PUBLISHED:
            proxy_model = versionables.for_content(content).version_model_proxy
            return reverse(
                f"admin:{proxy_model._meta.app_label}_{proxy_model.__name__.lower()}_edit_redirect",
                args=(version.pk,),
            )
    return ""

@register.filter
def url_revert_version(content, user):
    if hasattr(content, "prefetched_versions"):
        version = content.prefetched_versions[0]
    else:
        version = content.versions.first()
    if version:
        if version.check_revert.as_bool(user):
            proxy_model = versionables.for_content(content).version_model_proxy
            return reverse(
                f"admin:{proxy_model._meta.app_label}_{proxy_model.__name__.lower()}_revert",
                args=(version.pk,),
            )
    return ""



import typing

from django.conf import settings

from . import conf
from .exceptions import ConditionFailed
from .helpers import get_latest_draft_version, version_is_unlocked_for_user


class Conditions(list):
    def __add__(self, other: list) -> "Conditions":
        return Conditions(super().__add__(other))

    def __get__(self, instance: object, cls) -> typing.Union["Conditions", "BoundConditions"]:
        if instance:
            return BoundConditions(self, instance)
        return self

    def __call__(self, instance: object, user: settings.AUTH_USER_MODEL) -> None:
        for func in self:
            func(instance, user)

    def as_bool(self, instance: object, user: settings.AUTH_USER_MODEL) -> bool:
        try:
            self(instance, user)
        except ConditionFailed:
            return False
        return True


class BoundConditions:
    def __init__(self, conditions: Conditions, instance: object) -> None:
        self.conditions = conditions
        self.instance = instance

    def __call__(self, user) -> None:
        self.conditions(self.instance, user)

    def as_bool(self, user) -> bool:
        return self.conditions.as_bool(self.instance, user)


def in_state(states: list, message: str) -> callable:
    def inner(version, user):
        if version.state not in states:
            raise ConditionFailed(message)

    return inner


def is_not_locked(message: str) -> callable:
    """Condition that the version is not locked. Is only effective if ``settings.DJANGOCMS_VERSIONING_LOCK_VERSIONS``
    is set to ``True``"""
    def inner(version, user):
        if conf.LOCK_VERSIONS:
            if not version_is_unlocked_for_user(version, user):
                raise ConditionFailed(message.format(user=version.locked_by))
    return inner


def draft_is_not_locked(message: str) -> callable:
    def inner(version, user):
        if conf.LOCK_VERSIONS:
            draft_version = get_latest_draft_version(version)
            if draft_version and not version_is_unlocked_for_user(draft_version, user):
                raise ConditionFailed(message.format(user=draft_version.locked_by))
    return inner


def draft_is_locked(message: str) -> callable:
    def inner(version, user):
        if conf.LOCK_VERSIONS:
            draft_version = get_latest_draft_version(version)
            if not draft_version or version_is_unlocked_for_user(draft_version, user):
                raise ConditionFailed(message)
        else:
            raise ConditionFailed(message)
    return inner

def user_can_unlock(message: str) -> callable:
    def inner(version, user):
        if not user.has_perm("djangocms_versioning.delete_versionlock"):
            raise ConditionFailed(message)
    return inner

def user_can_publish(message: str) -> callable:
    def inner(version, user):
        if not version.has_publish_permission(user):
            raise ConditionFailed(message)
    return inner


def user_can_change(message: str) -> callable:
    def inner(version, user):
        if not version.has_change_permission(user):
            raise ConditionFailed(message)
    return inner



from datetime import datetime

from cms.api import add_plugin
from cms.models import Placeholder, UserSettings
from cms.test_utils.testcases import CMSTestCase
from freezegun import freeze_time

from djangocms_versioning.models import Version
from djangocms_versioning.test_utils import factories


class HandlersTestCase(CMSTestCase):
    def test_modified_date(self):
        pv = factories.PollVersionFactory()
        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            pv.content.save()
        pv = Version.objects.get(pk=pv.pk)
        self.assertEqual(pv.modified, dt)

    def test_add_plugin(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        placeholder.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI
        poll = factories.PollFactory()
        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_add_plugin_uri(
                placeholder=placeholder,
                plugin_type="PollPlugin",
                language=version.content.language,
            )
            data = {"poll": poll.pk}

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)
        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_change_plugin(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()

        plugin = add_plugin(
            placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_change_plugin_uri(plugin)
            data = {"poll": poll.pk}

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_clear_placeholder(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        placeholder.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_clear_placeholder_url(placeholder)

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, {"test": 0})
                self.assertEqual(response.status_code, 302)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_delete_plugin(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()

        plugin = add_plugin(
            placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_delete_plugin_uri(plugin)
            data = {"poll": poll.pk}

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 302)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_add_plugins_from_placeholder(self):
        version = factories.PageVersionFactory()
        source_placeholder = factories.PlaceholderFactory(source=version.content)
        target_placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()

        plugin = add_plugin(
            source_placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_copy_plugin_uri(plugin)
            data = {
                "source_language": version.content.language,
                "source_placeholder_id": source_placeholder.pk,
                "target_language": version.content.language,
                "target_placeholder_id": target_placeholder.pk,
            }

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_paste_placeholder(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()
        user_settings = UserSettings.objects.create(
            language=version.content.language,
            user=self.get_superuser(),
            clipboard=Placeholder.objects.create(slot="clipboard"),
        )

        placeholder_plugin = add_plugin(
            user_settings.clipboard, "PlaceholderPlugin", version.content.language
        )
        plugin = add_plugin(
            placeholder_plugin.placeholder_ref, "PollPlugin", version.content.language, poll=poll
        )

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_move_plugin_uri(plugin)
            data = {
                "plugin_id": placeholder_plugin.pk,
                "placeholder_id": placeholder.pk,
                "target_language": version.content.language,
                "move_a_copy": "true",
                "target_position": 1,
            }

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_paste_plugin(self):
        version = factories.PageVersionFactory()
        source_placeholder = factories.PlaceholderFactory(source=version.content)
        target_placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()

        plugin = add_plugin(
            source_placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"
        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_move_plugin_uri(plugin)
            data = {
                "plugin_id": plugin.pk,
                "placeholder_id": target_placeholder.pk,
                "target_language": version.content.language,
                "move_a_copy": "true",
                "target_position": 1,
            }

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_cut_plugin(self):
        version = factories.PageVersionFactory()
        placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()
        user_settings = UserSettings.objects.create(
            language=version.content.language,
            user=self.get_superuser(),
            clipboard=Placeholder.objects.create(slot="clipboard"),
        )

        plugin = add_plugin(
            placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_move_plugin_uri(plugin)
            data = {
                "plugin_id": plugin.pk,
                "target_language": version.content.language,
                "placeholder_id": user_settings.clipboard_id,
            }

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)

    def test_move_plugin(self):
        version = factories.PageVersionFactory()
        source_placeholder = factories.PlaceholderFactory(source=version.content)
        target_placeholder = factories.PlaceholderFactory(source=version.content)
        poll = factories.PollFactory()

        plugin = add_plugin(
            source_placeholder, "PollPlugin", version.content.language, poll=poll
        )
        plugin.page.get_absolute_url = lambda *args, **kwargs: "/test_page/"  # Fake URL needed for URI

        dt = datetime(2016, 6, 6)
        with freeze_time(dt):
            endpoint = self.get_move_plugin_uri(plugin)
            data = {
                "plugin_id": plugin.pk,
                "target_language": version.content.language,
                "placeholder_id": target_placeholder.pk,
                "target_position": 1,
            }

            with self.login_user_context(self.get_superuser()):
                response = self.client.post(endpoint, data)
                self.assertEqual(response.status_code, 200)

        version = Version.objects.get(pk=version.pk)
        self.assertEqual(version.modified, dt)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-23 14:05
from __future__ import unicode_literals

from django.db import migrations, models

import djangocms_versioning.models


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0013_auto_20181005_1404")]

    operations = [
        migrations.AddField(
            model_name="version",
            name="source",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=djangocms_versioning.models.allow_deleting_versions,
                to="djangocms_versioning.Version",
                verbose_name="source",
            ),
        )
    ]



import copy
import typing
import warnings
from contextlib import contextmanager

from cms.models import Page, PageContent, Placeholder
from cms.toolbar.utils import get_object_edit_url, get_object_preview_url
from cms.utils.helpers import is_editable_model
from cms.utils.urlutils import add_url_parameters, admin_reverse
from django.conf import settings
from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.mail import EmailMessage
from django.db import models
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.translation import get_language

from . import versionables
from .conf import EMAIL_NOTIFICATIONS_FAIL_SILENTLY
from .constants import DRAFT

try:
    from djangocms_internalsearch.helpers import emit_content_change
except ImportError:
    emit_content_change = None


def is_editable(content_obj, request):
    """Check of content_obj is editable"""
    from .models import Version

    return Version.objects.get_for_content(content_obj).check_modify.as_bool(
        request.user
    )


def versioning_admin_factory(admin_class, mixin):
    """A class factory returning admin class with overriden
    versioning functionality.

    :param admin_class: Existing admin class
    :param mixin: Mixin class
    :return: A subclass of `VersioningAdminMixin` and `admin_class`
    """
    return type("Versioned" + admin_class.__name__, (mixin, admin_class), {})


def _replace_admin_for_model(modeladmin, mixin, admin_site):
    """Replaces existing admin class registered for `modeladmin.model` with
    a subclass that includes versioning functionality.

    Doesn't do anything if `modeladmin` is already an instance of
    `mixin`.

    :param model: ModelAdmin instance
    :param mixin: Mixin class
    :param admin_site: AdminSite instance
    """
    if isinstance(modeladmin, mixin):
        return
    new_admin_class = versioning_admin_factory(modeladmin.__class__, mixin)
    admin_site.unregister(modeladmin.model)
    admin_site.register(modeladmin.model, new_admin_class)


def replace_admin_for_models(pairs, admin_site=None):
    """
    :param models: List of (model class, admin mixin class) tuples
    :param admin_site: AdminSite instance
    """
    if admin_site is None:
        admin_site = admin.site
    for model, mixin in pairs:
        try:
            modeladmin = admin_site._registry[model]
        except KeyError:
            continue
        _replace_admin_for_model(modeladmin, mixin, admin_site)


def register_versionadmin_proxy(versionable, admin_site=None):
    """Creates a model admin class based on `VersionAdmin` and registers
    it with `admin_site` for `versionable.version_model_proxy`.

    This model class applies filtering on the list of versions,
    so that only versions for `versionable.content_model` are shown.

    :param versionable: VersionableItem instance
    :param admin_site: AdminSite instance
    """
    from .admin import VersionAdmin

    if admin_site is None:
        admin_site = admin.site

    if versionable.version_model_proxy in admin_site._registry:
        # Attempting to register the proxy again is a no-op.
        warnings.warn(
            f"{versionable.version_model_proxy!r} is already registered with admin.",
            UserWarning,
            stacklevel=2,
        )
        return

    class VersionProxyAdminMixin(VersionAdmin):
        def get_queryset(self, request):
            return (
                super()
                .get_queryset(request)
                .filter(content_type__in=versionable.content_types)
            )

    ProxiedAdmin = type(
        versionable.grouper_model.__name__ + VersionAdmin.__name__,
        (VersionProxyAdminMixin, admin.ModelAdmin),
        {},
    )

    admin_site.register(versionable.version_model_proxy, ProxiedAdmin)


def manager_factory(manager, prefix, mixin):
    """A class factory returning a manager class with an added mixin to override for
    versioning functionality.

    :param manager: Existing manager class
    :return: A subclass of `PublishedContentManagerMixin` and `manager`
    """
    return type(
        prefix + manager.__name__,
        (mixin, manager),
        {"use_in_migrations": False},
    )


def replace_manager(model, manager, mixin, **kwargs):
    if hasattr(model, manager) and isinstance(getattr(model, manager), mixin):
        return
    original_manager = (
        getattr(model, manager).__class__ if hasattr(model, manager) else models.Manager
    )
    manager_object = manager_factory(original_manager, "Versioned", mixin)()
    for key, value in kwargs.items():
        setattr(manager_object, key, value)
    model._meta.local_managers = [
        mngr for mngr in model._meta.local_managers if mngr.name != manager
    ]
    model.add_to_class(manager, manager_object)
    if manager == "objects":
        # only safe the original default manager
        model.add_to_class(
            f'_original_{"manager" if manager == "objects" else manager}',
            original_manager(),
        )


def inject_generic_relation_to_version(model):
    from .models import Version

    related_query_name = f"{model._meta.app_label}_{model._meta.model_name}"
    model.add_to_class(
        "versions", GenericRelation(Version, related_query_name=related_query_name)
    )
    if not hasattr(model, "is_editable"):
        model.add_to_class("is_editable", is_editable)


def _set_default_manager(model, manager):
    model._meta.local_managers = [
        m for m in model._meta.local_managers if m.name != "objects"
    ]
    manager_ = copy.copy(manager)
    manager_.name = "objects"
    model.add_to_class("objects", manager_)


@contextmanager
def override_default_manager(model, manager):
    original_manager = model.objects
    _set_default_manager(model, manager)
    yield
    _set_default_manager(model, original_manager)


@contextmanager
def nonversioned_manager(model):
    manager_cls = model.objects.__class__
    manager_cls.versioning_enabled = False
    yield
    manager_cls.versioning_enabled = True


def _version_list_url(versionable, **params):
    proxy = versionable.version_model_proxy
    return add_url_parameters(
        admin_reverse(f"{proxy._meta.app_label}_{proxy._meta.model_name}_changelist"),
        **params,
    )


def version_list_url(content):
    """Returns a URL to list of content model versions,
    filtered by `content`'s grouper
    """
    versionable = versionables._cms_extension().versionables_by_content[
        content.__class__
    ]
    return _version_list_url(
        versionable, **versionable.grouping_values(content, relation_suffix=False)
    )


def version_list_url_for_grouper(grouper):
    """Returns a URL to list of content model versions,
    filtered by `grouper`
    """
    versionable = versionables._cms_extension().versionables_by_grouper[
        grouper.__class__
    ]
    return _version_list_url(
        versionable, **{versionable.grouper_field_name: str(grouper.pk)}
    )


def is_content_editable(placeholder, user):
    """A helper method for monkey patch to check version is in edit state.
    Returns True if placeholder is related to a source object
    which is not versioned.

    :param placeholder: current placeholder
    :param user: user object
    :return: Boolean
    """
    try:
        versionables.for_content(placeholder.source)
    except KeyError:
        return True
    from .models import Version

    version = Version.objects.get_for_content(placeholder.source)
    return version.state == DRAFT


def get_editable_url(content_obj, force_admin=False):
    """If the object is editable the cms editable view should be used, with the toolbar.
    This method provides the URL for it.
    """
    if is_editable_model(content_obj.__class__) and not force_admin:
        language = getattr(content_obj, "language", None)
        url = get_object_edit_url(content_obj, language)
    # Or else, the standard edit view should be used
    else:
        url = admin_reverse(
            f"{content_obj._meta.app_label}_{content_obj._meta.model_name}_change",
            args=(content_obj.pk,),
        )
    return url


# TODO Based on polymorphic.query_translate._get_mro_content_type_ids,
# can use that when polymorphic gets a new release
def get_content_types_with_subclasses(models, using=None):
    content_types = set()
    for model in models:
        content_type = ContentType.objects.db_manager(using).get_for_model(
            model, for_concrete_model=False
        )
        content_types.add(content_type.pk)
        subclasses = model.__subclasses__()
        if subclasses:
            content_types.update(get_content_types_with_subclasses(subclasses, using))
    return content_types


def get_preview_url(
    content_obj: models.Model, language: typing.Union[str, None] = None
) -> str:
    """If the object is editable the cms preview view should be used, with the toolbar.
    This method provides the URL for it. It falls back the standard change view
    should the object not be frontend editable.
    """
    versionable = versionables.for_content(content_obj)
    if versionable.preview_url:
        return versionable.preview_url(content_obj)
    if is_editable_model(content_obj.__class__):
        if not language:
            # Use language field is content object has one to determine the language
            language = getattr(content_obj, "language", get_language())
        url = get_object_preview_url(content_obj, language=language)
    else:
        # Or else, the standard change view should be used
        url = admin_reverse(
            f"{content_obj._meta.app_label}_{content_obj._meta.model_name}_change",
            args=[content_obj.pk],
        )
        if language:
            url += f"&language={language}"
    return url


def get_admin_url(model: type, action: str, *args) -> str:
    opts = model._meta
    url_name = f"{opts.app_label}_{opts.model_name}_{action}"
    return admin_reverse(url_name, args=args)


def remove_published_where(queryset):
    """
    By default, the versioned queryset filters out so that only versions
    that are published are returned. If you need to return the full queryset
    use the "admin_manager" instead of "objects"
    """
    raise NotImplementedError(
        "remove_published_where has been replaced by ContentObj.admin_manager"
    )


def get_latest_admin_viewable_content(
    grouper: models.Model,
    include_unpublished_archived: bool = False,
    **extra_grouping_fields,
) -> models.Model:
    """
    Return the latest Draft or Published PageContent using the draft where possible
    """
    versionable = versionables.for_grouper(grouper)

    # Check if all required grouping fields are given to be able to select the latest admin viewable content
    missing_fields = [
        field
        for field in versionable.extra_grouping_fields
        if field not in extra_grouping_fields
    ]
    if missing_fields:
        raise ValueError(
            f"Grouping field(s) {missing_fields} required for {versionable.grouper_model}."
        )

    # Get the name of the content_set (e.g., "pagecontent_set") from the versionable
    content_set = versionable.grouper_field.remote_field.get_accessor_name()

    # Accessing the content set through the grouper preserves prefetches
    qs = getattr(grouper, content_set)(manager="admin_manager")

    if include_unpublished_archived:
        # Relevant for admin to see e.g., the latest unpublished or archived versions
        return qs.filter(**extra_grouping_fields).latest_content().first()
    # Return only active versions, e.g., for copying
    return qs.filter(**extra_grouping_fields).current_content().first()


def get_latest_admin_viewable_page_content(
    page: Page, language: str
) -> PageContent:  # pragma: no cover
    warnings.warn(
        "get_latst_admin_viewable_page_content has ben deprecated. "
        "Use get_latest_admin_viewable_content(page, language=language) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_latest_admin_viewable_content(page, language=language)


def proxy_model(obj: models.Model, content_model: type) -> models.Model:
    """
    Get the proxy model from a

    :param obj: A registered versionable object
    :param content_model: A registered content model
    """
    versionable = versionables.for_content(content_model)
    obj_ = copy.deepcopy(obj)
    obj_.__class__ = versionable.version_model_proxy
    return obj_


def create_version_lock(version, user):
    """
    Create a version lock if necessary
    """
    changed = version.locked_by != user
    version.locked_by = user
    version.save()
    if changed and emit_content_change:
        emit_content_change(version.content)
    return version


def remove_version_lock(version):
    """
    Delete a version lock, handles when there are none available.
    """
    return create_version_lock(version, None)


def version_is_locked(version) -> settings.AUTH_USER_MODEL:
    """
    Determine if a version is locked
    """
    return version.locked_by


def version_is_unlocked_for_user(version, user: settings.AUTH_USER_MODEL) -> bool:
    """Check if lock doesn't exist for a version object or is locked to provided user."""
    return version.locked_by is None or version.locked_by == user


def content_is_unlocked_for_user(
    content: models.Model, user: settings.AUTH_USER_MODEL
) -> bool:
    """Check if lock doesn't exist or object is locked to provided user."""
    try:
        if hasattr(content, "prefetched_versions"):
            version = content.prefetched_versions[0]
        else:
            version = content.versions.first()
        return version_is_unlocked_for_user(version, user)
    except AttributeError:
        return True


def placeholder_content_is_unlocked_for_user(
    placeholder: Placeholder, user: settings.AUTH_USER_MODEL
) -> bool:
    """Check if lock doesn't exist or placeholder source object
    is locked to provided user.
    """
    content = placeholder.source
    return content_is_unlocked_for_user(content, user)


def send_email(
    recipients: list, subject: str, template: str, template_context: dict
) -> int:
    """
    Send emails using locking templates
    """
    template = f"djangocms_versioning/emails/{template}"
    subject = force_str(subject)
    content = render_to_string(template, template_context)

    message = EmailMessage(
        subject=subject,
        body=content,
        from_email=settings.DEFAULT_FROM_EMAIL,
        to=recipients,
    )
    return message.send(fail_silently=EMAIL_NOTIFICATIONS_FAIL_SILENTLY)


def get_latest_draft_version(version: models.Model) -> models.Model:
    """Get latest draft version of version object and caches it in the
    content object"""
    from .models import Version

    if (
        not hasattr(version.content, "_latest_draft_version")
        or getattr(version.content._latest_draft_version, "state", DRAFT) != DRAFT
    ):
        drafts = Version.objects.filter_by_content_grouping_values(
            version.content
        ).filter(state=DRAFT)
        version.content._latest_draft_version = drafts.first()
    return version.content._latest_draft_version



from django.db import models
from django.urls import reverse


class BlogPost(models.Model):
    name = models.TextField()

    def __str__(self):
        return f"{self.name} ({self.pk})"


class BlogContent(models.Model):
    blogpost = models.ForeignKey(BlogPost, on_delete=models.CASCADE)
    language = models.TextField()
    text = models.TextField()

    def has_publish_permission(self, user):
        if user.is_superuser:
            return True
        # Fake a simple object-dependent permission
        return user.username in self.text

    def has_change_permission(self, user):
        if user.is_superuser:
            return True
        # Fake a simple object-dependent permission
        return f"<{user.username}>" in self.text

    def __str__(self):
        return self.text

    def get_absolute_url(self):
        return reverse("admin:blogpost_blogcontent_changelist")


class Comment(models.Model):
    blogpost = models.ForeignKey(BlogPost, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.pk)


class CommentContent(models.Model):
    comment = models.ForeignKey(Comment, on_delete=models.CASCADE)
    text = models.TextField()

    def __str__(self):
        return self.text

    def get_absolute_url(self):
        return reverse("admin:blogpost_commentcontent_changelist")



import json
import typing
import warnings
from collections import OrderedDict
from urllib.parse import urlparse

from cms.admin.utils import CONTENT_PREFIX, ChangeListActionsMixin, GrouperModelAdmin
from cms.models import PageContent
from cms.utils import get_language_from_request
from cms.utils.conf import get_cms_setting
from cms.utils.helpers import is_editable_model
from cms.utils.urlutils import add_url_parameters, static_with_version
from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin.actions import delete_selected
from django.contrib.admin.options import IncorrectLookupParameters
from django.contrib.admin.utils import unquote
from django.contrib.admin.views.main import ChangeList
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist, PermissionDenied
from django.db import models
from django.db.models import OuterRef, Subquery
from django.db.models.functions import Cast, Lower
from django.forms import MediaDefiningClass
from django.http import (
    Http404,
    HttpRequest,
    HttpResponseForbidden,
    HttpResponseNotAllowed,
)
from django.shortcuts import redirect, render
from django.template.loader import render_to_string, select_template
from django.template.response import TemplateResponse
from django.urls import Resolver404, path, resolve, reverse
from django.utils.encoding import force_str
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from . import conf, versionables
from .constants import DRAFT, INDICATOR_DESCRIPTIONS, PUBLISHED, VERSION_STATES
from .emails import notify_version_author_version_unlocked
from .exceptions import ConditionFailed
from .forms import grouper_form_factory
from .helpers import (
    content_is_unlocked_for_user,
    create_version_lock,
    get_admin_url,
    get_editable_url,
    get_latest_admin_viewable_content,
    get_preview_url,
    proxy_model,
    remove_version_lock,
    version_is_locked,
    version_list_url,
)
from .indicators import content_indicator, content_indicator_menu
from .models import Version
from .versionables import _cms_extension


class VersioningChangeListMixin:
    """Mixin used for ChangeList classes of content models."""

    def get_queryset(self, request, exclude_parameters=None):
        """Limit the content model queryset to the latest versions only."""
        if exclude_parameters:
            # Django 5.0+ (facet support)
            queryset = super().get_queryset(request, exclude_parameters)
        else:
            # Django 4.2 compatible get_queryset
            queryset = super().get_queryset(request)
        versionable = versionables.for_content(queryset.model)

        """Check if there is a method "self.get_<field>_from_request" for each extra grouping field.
         If so call it to retrieve the appropriate filter. If no method is found (except for "language")
         no filter is applied. For "language" the fallback is versioning's "get_language_frmo_request".

         Admins requiring extra grouping field beside "language" need to implement the "get_<field>_from_request"
         method themselves. A common way to select the field might be GET or POST parameters or user-related settings.
         """

        grouping_filters = {}
        for field in versionable.extra_grouping_fields:
            if hasattr(self.model_admin, f"get_{field}_from_request"):
                grouping_filters[field] = getattr(self.model_admin, f"get_{field}_from_request")(request)
            elif field == "language":
                grouping_filters[field] = get_language_from_request(request)
        return queryset.filter(pk__in=versionable.distinct_groupers(**grouping_filters))


def versioning_change_list_factory(base_changelist_cls):
    """Generate a ChangeList class to use for the content model"""
    return type(
        "Versioned" + base_changelist_cls.__name__,
        (VersioningChangeListMixin, base_changelist_cls),
        {}
    )


class VersioningAdminMixin:
    """Mixin providing versioning functionality to admin classes of
    content models.
    """

    def save_model(self, request, obj, form, change):
        """
        Overrides the save method to create a version object
        when a content object is created
        """
        super().save_model(request, obj, form, change)
        if not change:
            if isinstance(obj, PageContent):
                # FIXME disabled version creation for `cms.PageContent`
                # here, as it's already done in `cms.api.create_title`
                return
            # create a new version object and save it
            Version.objects.create(content=obj, created_by=request.user)

    def get_queryset(self, request):
        """Override manager so records not in published state can be displayed"""
        from .helpers import override_default_manager

        with override_default_manager(self.model, self.model._original_manager):
            queryset = super().get_queryset(request)
        return queryset

    def get_changelist(self, request, **kwargs):
        ChangeList = super().get_changelist(request, **kwargs)
        return versioning_change_list_factory(ChangeList)

    change_form_template = "djangocms_versioning/admin/mixin/change_form.html"

    def render_change_form(
        self, request, context, add=False, change=False, form_url="", obj=None
    ):
        """Add a link to the version table to the change form view"""
        if "versioning_fallback_change_form_template" not in context:
            context[
                "versioning_fallback_change_form_template"
            ] = super().change_form_template

        return super().render_change_form(request, context, add=add, change=change, form_url=form_url, obj=obj)

    def has_change_permission(self, request, obj=None):
        # Add additional version checks
        if obj:
            version = Version.objects.get_for_content(obj)
            permission = version.check_modify.as_bool(request.user)
            if conf.LOCK_VERSIONS and permission:
                permission = content_is_unlocked_for_user(obj, request.user)
            return permission

        return super().has_change_permission(request, obj)


class StateIndicatorMixin(metaclass=MediaDefiningClass):
    """Mixin to provide state_indicator column to the changelist view of a content model admin. Usage::

        class MyContentModelAdmin(StateIndicatorMixin, admin.ModelAdmin):
            list_display = [..., "state_indicator", ...]
    """
    class Media:
        # js for the context menu
        js = ("admin/js/jquery.init.js", "djangocms_versioning/js/indicators.js",)
        # css for indicators and context menu
        css = {
            "all": (static_with_version("cms/css/cms.pagetree.css"),),
        }

    indicator_column_label = _("State")

    @property
    def _extra_grouping_fields(self):
        try:
            return versionables.for_grouper(self.model).extra_grouping_fields
        except KeyError:
            return None

    def get_indicator_column(self, request):
        def indicator(obj):
            if self._extra_grouping_fields is not None:  # Grouper Model
                content_obj = get_latest_admin_viewable_content(obj, include_unpublished_archived=True, **{
                    field: getattr(self, field) for field in self._extra_grouping_fields
                })
            else:  # Content Model
                content_obj = obj
            status = content_indicator(content_obj)
            menu = content_indicator_menu(
                request,
                status,
                content_obj._version,
                back=request.path_info + "?" + request.GET.urlencode(),
            ) if status else None
            return render_to_string(
                "admin/djangocms_versioning/indicator.html",
                {
                    "state": status or "empty",
                    "description": INDICATOR_DESCRIPTIONS.get(status, _("Empty")),
                    "menu_template": "admin/cms/page/tree/indicator_menu.html",
                    "menu": json.dumps(render_to_string("admin/cms/page/tree/indicator_menu.html",
                                                        {"indicator_menu_items": menu})) if menu else None,
                }
            )
        indicator.short_description = self.indicator_column_label
        return indicator

    def state_indicator(self, obj):
        raise ValueError(
            'ModelAdmin.display_list contains "state_indicator" as a placeholder for status indicators. '
            'Status indicators, however, are not loaded. If you implement "get_list_display" make '
            "sure it calls super().get_list_display."
        )  # pragma: no cover

    def get_list_display(self, request):
        """Default behavior: replaces the text "state_indicator" by the indicator column"""
        if versionables.exists_for_content(self.model) or versionables.exists_for_grouper(self.model):
            return tuple(self.get_indicator_column(request) if item == "state_indicator" else item
                         for item in super().get_list_display(request))
        else:
            # remove "state_indicator" entry
            return tuple(item for item in super().get_list_display(request) if item != "state_indicator")


class ExtendedListDisplayMixin:
    """Implements the extend_list_display method at allows other packages to add fields to the list display
    of a verisoned object"""

    @property
    def _is_grouper_admin(self):
        return isinstance(self, GrouperModelAdmin)

    def _get_field_modifier(self, request, modifier_dict, field):
        method = modifier_dict[field]

        def get_field_modifier(obj):
            if self._is_grouper_admin:  # In a grouper admin?
                return method(self.get_content_obj(obj), field)
            else:
                return method(obj, field)

        get_field_modifier.short_description = field
        return get_field_modifier

    def extend_list_display(self, request, modifier_dict, list_display):
        list_display = [*list_display]
        for field in modifier_dict:
            if not callable(modifier_dict[field]):
                raise ImproperlyConfigured("Field provided must be callable")
            try:
                prefix = CONTENT_PREFIX if self._is_grouper_admin else ""
                field_modifier = self._get_field_modifier(request, modifier_dict, field)
                list_display[list_display.index(prefix + field)] = field_modifier
            except ValueError:
                raise ImproperlyConfigured("The target field does not exist in this context") from None
        return tuple(list_display)

    def get_list_display(self, request):
        # get configured list_display
        list_display = super().get_list_display(request)
        # Get the versioning extension
        extension = _cms_extension()
        if isinstance(self, GrouperModelAdmin):
            modifier_dict = extension.add_to_field_extension.get(self.content_model, None)
        else:
            modifier_dict = extension.add_to_field_extension.get(self.model, None)
        if modifier_dict:
            list_display = self.extend_list_display(request, modifier_dict, list_display)
        return list_display


class ExtendedGrouperVersionAdminMixin(ExtendedListDisplayMixin):
    """Mixin to provide state_indicator, author and changed date column to the changelist view of a
    grouper model admin. Usage::

        class MyContentModelAdmin(ExtendedGrouperVersionAdminMixin, cms.admin.utils.GrouperModelAdmin):
            list_display = [
                ...,
                "get_author",   # Adds the author column
                "get_modified_date",  # Adds the modified column
                "get_versioning_state",  # Adds the state (w/o interaction)
                ...]

        """
    def get_queryset(self, request: HttpRequest) -> models.QuerySet:
        """Annotates the username of the ``created_by`` field, the ``modified`` field (date time),
        and the ``state`` field of the version object to the grouper queryset."""
        grouper_content_type = versionables.for_grouper(self.model).content_types
        qs = super().get_queryset(request)
        versions = Version.objects.filter(object_id=OuterRef("pk"), content_type__in=grouper_content_type)
        contents = self.content_model.admin_manager.latest_content(
            **{self.grouper_field_name: OuterRef("pk"), **self.current_content_filters}
        ).annotate(
            content_created_by=Subquery(versions.values(f"created_by__{conf.USERNAME_FIELD}")[:1]),
            content_state=Subquery(versions.values("state")),
            content_modified=Subquery(versions.values("modified")[:1]),
        )
        qs = qs.annotate(
            content_created_by=Subquery(contents.values("content_created_by")[:1]),
            content_created_by_sort=Lower(Subquery(contents.values("content_created_by")[:1])),
            content_state=Subquery(contents.values("content_state")),
            # cast is necessary for mysql
            content_modified=Cast(Subquery(contents.values("content_modified")[:1]), models.DateTimeField()),
        )
        return qs

    @admin.display(
        description=_("State"),
        ordering="content_state",
    )
    def get_versioning_state(self, obj: models.Model) -> typing.Union[str, None]:
        """Returns verbose text of objects versioning state. This is a text column without user interaction.
        Typically, either ``get_versioning_state`` or ``state_indicator`` (provided by the
        :class:`~djangocms_versioning.admin.StateIndicatorMixin`) is used. The state indicator
        allows for user interaction.
        :param obj: Versioned grouper model instance annotated with its content state
        :return: description of state
        """
        return dict(VERSION_STATES).get(obj.content_state)

    @admin.display(
        description=_("Author"),
        ordering="content_created_by_sort",
    )
    def get_author(self, obj: models.Model) -> typing.Union[str, None]:
        """
        Return the author who created a version
        :param obj: Versioned grouper model instance annotated with its author username
        :return: Author username
        """
        return getattr(obj, "content_created_by", None)

    # This needs to target the annotation, or ordering will be alphabetically, with uppercase then lowercase

    @admin.display(
        description=_("Modified"),
        ordering="content_modified",
    )
    def get_modified_date(self, obj: models.Model) -> typing.Union[str, None]:
        """
        Get the last modified date of a version
        :param obj: Versioned grouper model instance annotated with its modified datetime
        :return: Modified Date
        """
        return getattr(obj, "content_modified", None)


class ExtendedVersionAdminMixin(
    ExtendedListDisplayMixin,
    ChangeListActionsMixin,
    VersioningAdminMixin,
    metaclass=MediaDefiningClass,
):
    """
    Extended VersionAdminMixin for common/generic versioning admin items

    CAVEAT: Ordered fields are implemented by this mixin, if custom ordering is added to any models that
    inherits this Mixin it will require accommodating/reimplementing this.
    """

    versioning_list_display = (
        "get_author",
        "get_modified_date",
        "get_versioning_state",
    )

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        # Due to django admin ordering using unicode, to alphabetically order regardless of case, we must
        # annotate the queryset, with the usernames all lower case, and then order based on that!

        queryset = queryset.annotate(created_by_username_ordering=Lower(f"versions__created_by__{conf.USERNAME_FIELD}"))
        return queryset

    def get_version(self, obj):
        """
        Return the latest version of a given object
        :param obj: Versioned Content instance
        :return: Latest Version linked with content instance
        """
        return obj.versions.all()[0]

    @admin.display(
        description=_("State"),
        ordering="versions__state",
    )
    def get_versioning_state(self, obj):
        """
        Return the state of a given version
        """
        return self.get_version(obj).get_state_display()

    @admin.display(
        description=_("Author"),
        ordering="created_by_username_ordering",
    )
    def get_author(self, obj):
        """
        Return the author who created a version
        :param obj: Versioned content model Instance
        :return: Author
        """
        return self.get_version(obj).created_by

    # This needs to target the annotation, or ordering will be alphabetically, with uppercase then lowercase

    @admin.display(
        description=_("Modified"),
        ordering="versions__modified",
    )
    def get_modified_date(self, obj):
        """
        Get the last modified date of a version
        :param obj: Versioned content model Instance
        :return: Modified Date
        """
        return self.get_version(obj).modified

    def _get_preview_url(self, obj):
        """
        Return the preview method if available, otherwise return None
        :return: method or None
        """
        if hasattr(obj, "get_preview_url"):
            return obj.get_preview_url()
        else:
            return None

    def _get_preview_link(self, obj, request, disabled=False):
        """
        Return a user-friendly button for previewing the content model
        :param obj: Instance of versioned content model
        :param request: The request to admin menu
        :param disabled: Should the link be marked disabled?
        :return: Preview icon template
        """
        preview_url = self._get_preview_url(obj) or get_preview_url(obj)
        if not preview_url:
            disabled = True

        return self.admin_action_button(
            preview_url,
            icon="view",
            title=_("Preview"),
            name="preview",
            keepsideframe=False,
            disabled=disabled,
        )

    def _get_edit_link(self, obj, request, disabled=False):
        """
        Return a user-friendly button for editing the content model
        - mark disabled if user doesn't have permission
        - hide completely if instance cannot be edited
        :param obj: Instance of Versioned model
        :param request: The request to admin menu
        :param disabled: Should the link be marked disabled?
        :return: Preview icon template
        """
        version = proxy_model(self.get_version(obj), self.model)

        if not version.check_edit_redirect.as_bool(request.user):
            # Don't display the link if it can't be edited
            return ""

        if not request.user.has_perm(f"{obj._meta.app_label}.{obj._meta.model_name}"):
            # Grey out if user has not sufficient right to edit
            disabled = True

        url = reverse(
            f"admin:{version._meta.app_label}_{version._meta.model_name}_edit_redirect",
            args=(version.pk,),
        )
        # Only show if no draft exists
        if version.state == PUBLISHED:
            pks_for_grouper = version.versionable.for_content_grouping_values(
                obj
            ).values_list("pk", flat=True)
            drafts = Version.objects.filter(
                object_id__in=pks_for_grouper,
                content_type=version.content_type,
                state=DRAFT,
            )
            if drafts.exists():
                return ""
            icon = "edit-new"
        else:
            icon = "edit"

        return self.admin_action_button(
            url,
            icon=icon,
            title=_("Edit") if icon == "edit" else _("New Draft"),
            name="edit",
            disabled=disabled,
            action="post",
            keepsideframe=False,
        )

    def _get_manage_versions_link(self, obj, request, disabled=False):
        url = version_list_url(obj)
        return self.admin_action_button(
            url,
            icon="copy",
            title=_("Manage versions"),
            name="manage-versions",
            disabled=disabled,
        )

    def get_actions_list(self):
        """
        Collect rendered actions from implemented methods and return as list
        """
        actions = [
            self._get_preview_link,
            self._get_edit_link,
        ]
        if "state_indicator" not in self.versioning_list_display:
            # State indicator mixin loaded?
            actions.append(self._get_manage_versions_link)
        return actions

    def get_list_display(self, request):
        # get configured list_display
        list_display = super().get_list_display(request)
        # Add versioning information and action fields
        list_display += self.versioning_list_display + (self.get_admin_list_actions(request),)
        return list_display


class ExtendedIndicatorVersionAdminMixin(StateIndicatorMixin, ExtendedVersionAdminMixin):
    versioning_list_display = (
        "get_author",
        "get_modified_date",
        "state_indicator",
    )


class VersionChangeList(ChangeList):
    def get_filters_params(self, params=None):
        """Removes the grouper param from the filters as the main grouper
        filtering is not handled by the UI filters and therefore needs to be
        handled differently.
        """
        content_model = self.model_admin.model._source_model
        versionable = versionables.for_content(content_model)
        filter_params = super().get_filters_params(params)
        filter_params.pop(versionable.grouper_field_name, None)
        return filter_params

    def get_grouping_field_filters(self, request):
        """Handles extra grouping params (such as PageContent.language).

        The get_filters_params method does return these filters as they are
        visible in the UI, however they need extra handling due to db
        optimization and the difficulties involved in handling the
        generic foreign key from Version to the content model."""
        content_model = self.model_admin.model._source_model
        versionable = versionables.for_content(content_model)
        fields = versionable.grouping_fields
        for field in fields:
            value = request.GET.get(field)
            if value is not None:
                yield field, value

    def get_queryset(self, request, exclude_parameters=None):
        """Adds support for querying the version model by grouping fields.

        Filters by the value of grouping fields (specified in VersionableItem
        definition) of content model.

        Functionality is implemented here, because list_filter doesn't allow
        for specifying filters that work without being shown in the UI
        along with filter choices.
        """
        if exclude_parameters:
            # Django 5.0+ (facet support)
            queryset = super().get_queryset(request, exclude_parameters)
        else:
            # Django 4.2 compatible get_queryset
            queryset = super().get_queryset(request)
        content_model = self.model_admin.model._source_model
        versionable = versionables.for_content(content_model)
        filters = dict(self.get_grouping_field_filters(request))
        if versionable.grouper_field_name not in filters:
            raise IncorrectLookupParameters("Missing grouper")
        return queryset.filter_by_grouping_values(versionable, **filters)


def fake_filter_factory(versionable, field_name):
    """Returns filters that merely expose the filtering UI,
    without having any effect on the resulting queryset.
    """
    field = versionable.content_model._meta.get_field(field_name)
    lookups_ = versionable.version_list_filter_lookups[field_name]

    class FakeFilter(admin.SimpleListFilter):
        title = field.verbose_name
        parameter_name = field_name

        def lookups(self, request, model_admin):
            if callable(lookups_):
                return lookups_()
            else:
                return lookups_

        def queryset(self, request, queryset):
            return queryset

    return FakeFilter


class VersionAdmin(ChangeListActionsMixin, admin.ModelAdmin, metaclass=MediaDefiningClass):
    """Admin class used for version models.
    """

    # register custom actions
    actions = ["compare_versions", "delete_selected"]
    list_display = (
        "number",
    ) + (
        ("created",) if conf.VERBOSE_UI else ()
    ) + (
        "modified",
        "content",
        "created_by",
    ) + (
        ("locked",) if conf.LOCK_VERSIONS else ()
    ) + (
        "state",
        "admin_list_actions",
    )
    list_display_links = None

    # FIXME disabled until GenericRelation attached to content models gets
    # fixed to include subclass (polymorphic) support
    #
    # def get_queryset(self, request):
    #     return super().get_queryset(request).prefetch_related('content')

    class Media:
        js = ["djangocms_versioning/js/versioning.js"]

    def has_module_permission(self, request):
        return conf.VERBOSE_UI

    def get_changelist(self, request, **kwargs):
        return VersionChangeList

    def get_list_filter(self, request):
        """Adds the filters for the extra grouping fields to the UI."""
        versionable = versionables.for_content(self.model._source_model)
        return [
            fake_filter_factory(versionable, field)
            for field in versionable.extra_grouping_fields
        ]

    @admin.display(
        description=_("Content"),
        ordering="content",
    )
    def content_link(self, obj):
        """Display html for the content preview url - replaced by Preview action"""
        warnings.warn("VersionAdmin.content_link is deprecated.", DeprecationWarning, stacklevel=2)
        content = obj.content
        url = get_preview_url(content)

        return format_html(
            '<a target="_top" class="js-close-sideframe" href="{url}">{label}</a>',
            url=mark_safe(url),
            label=content,
        )

    @admin.display(
        description=_("locked")
    )
    def locked(self, version):
        """
        Generate an locked field for Versioning Admin
        """
        if version.state == DRAFT and version_is_locked(version):
            return mark_safe('<span class="cms-icon cms-icon-lock"></span>')
        return ""

    def _get_preview_link(self, obj, request):
        if obj.state == DRAFT:
            # Draft versions have edit button
            return ""
        url = get_preview_url(obj.content)
        return self.admin_action_button(
            url,
            icon="view",
            name="preview",
            keepsideframe=False,
            title=_("Preview"),
        )

    def _get_archive_link(self, obj, request, disabled=False):
        """Helper function to get the html link to the archive action
        """
        if not obj.can_be_archived():
            # Don't display the link if it can't be archived
            return ""
        archive_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_archive",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            archive_url,
            icon="archive",
            title=_("Archive"),
            name="archive",
            disabled=not obj.check_archive.as_bool(request.user),
        )

    def _get_publish_link(self, obj, request):
        """Helper function to get the html link to the publish action
        """
        if not obj.can_be_published():
            # Don't display the link if it can't be published
            return ""
        publish_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_publish",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            publish_url,
            icon="publish",
            title=_("Publish"),
            name="publish",
            action="post",
            disabled=not obj.check_publish.as_bool(request.user),
            keepsideframe=False,
        )

    def _get_unpublish_link(self, obj, request, disabled=False):
        """Helper function to get the html link to the unpublish action
        """
        if not obj.can_be_unpublished():
            # Don't display the link if it can't be unpublished
            return ""
        unpublish_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_unpublish",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            unpublish_url,
            icon="unpublish",
            title=_("Unpublish"),
            name="unpublish",
            disabled=not obj.check_unpublish.as_bool(request.user),
        )

    def _get_edit_link(self, obj, request, disabled=False):
        """Helper function to get the html link to the edit action
        """

        if not obj.check_edit_redirect.as_bool(request.user):
            return ""

        # Only show if no draft exists
        if obj.state == PUBLISHED:
            pks_for_grouper = obj.versionable.for_content_grouping_values(
                obj.content
            ).values_list("pk", flat=True)
            drafts = Version.objects.filter(
                object_id__in=pks_for_grouper,
                content_type=obj.content_type,
                state=DRAFT,
            )
            if drafts.exists():
                return ""
            icon = "edit-new"
        else:
            icon = "edit"

        # Don't open in the sideframe if the item is not sideframe compatible
        keepsideframe = obj.versionable.content_model_is_sideframe_editable

        edit_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_edit_redirect",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            edit_url,
            icon=icon,
            title=_("Edit") if icon == "edit" else _("New Draft"),
            name="edit",
            action="post",
            disabled=disabled,
            keepsideframe=keepsideframe,
        )

    def _get_revert_link(self, obj, request, disabled=False):
        """Helper function to get the html link to the revert action
        """
        if obj.state in (PUBLISHED, DRAFT):
            # Don't display the link if it's a draft or published
            return ""

        revert_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_revert",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            revert_url,
            icon="undo",
            title=_("Revert"),
            name="revert",
            disabled=not obj.check_revert.as_bool(request.user) or disabled,
        )

    def _get_discard_link(self, obj, request, disabled=False):
        """Helper function to get the html link to the discard action
        """
        if obj.state != DRAFT:
            # Don't display the link if it's not a draft
            return ""

        discard_url = reverse(
            f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_discard",
            args=(obj.pk,),
        )
        return self.admin_action_button(
            discard_url,
            icon="bin",
            title=_("Discard"),
            name="discard",
            disabled=not obj.check_discard.as_bool(request.user) or disabled,
        )

    def _get_unlock_link(self, obj, request):
        """
        Generate an unlock link for the Versioning Admin
        """
        # If the version is not draft no action should be present
        if not conf.LOCK_VERSIONS or obj.state != DRAFT or not version_is_locked(obj):
            return ""

        unlock_url = reverse(f"admin:{obj._meta.app_label}_{self.model._meta.model_name}_unlock", args=(obj.pk,))
        return self.admin_action_button(
            unlock_url,
            icon="unlock",
            title=_("Unlock"),
            name="unlock",
            action="post",
            disabled=not obj.check_unlock.as_bool(request.user),
        )

    def _get_settings_link(self, obj, request):
        """
        Generate a settings button for the Versioning Admin
        """

        # If the content object is not registered for frontend editing no action should be present
        # Also, the content object must be registered with the admin site
        content_model = obj.versionable.content_model
        if not is_editable_model(content_model):
            return ""

        try:
            settings_url = reverse(
                f"admin:{content_model._meta.app_label}_{content_model._meta.model_name}_change",
                args=(obj.content.pk,)
            )
        except Resolver404:
            return ""

        return self.admin_action_button(
            settings_url,
            icon="settings",
            title=_("Settings"),
            name="settings",
        )

    def get_actions_list(self):
        """Returns all action links as a list"""
        return self.get_state_actions()

    def get_state_actions(self):
        """Compatibility shim for djangocms-moderation. Do not use.
        It will be removed in a future version."""

        if settings.DEBUG:
            # Only introspect in DEBUG mode. Issue warning if method is monkey-patched
            import inspect
            caller_frame = inspect.getouterframes(inspect.currentframe(), 2)
            if caller_frame[1][3] != "get_actions_list":
                warnings.warn("Modifying get_state_actions is deprecated. VersionAdmin.get_state_actions "
                              "will be removed in a future version. Use get_actions_list instead.",
                              DeprecationWarning, stacklevel=2)

        return [
            self._get_preview_link,
            self._get_edit_link,
            self._get_archive_link,
            self._get_publish_link,
            self._get_unpublish_link,
            self._get_revert_link,
            self._get_discard_link,
            self._get_unlock_link,
            self._get_settings_link,
        ]

    @admin.action(
        description=_("Compare versions")
    )
    def compare_versions(self, request, queryset):
        """
        Redirects to a compare versions view based on a users choice
        """
        queryset = queryset.order_by("pk")

        # Validate that only two versions are selected
        if queryset.count() != 2:
            self.message_user(request, _("Exactly two versions need to be selected."))
            return

        # Build the link for the version comparison of the two selected versions
        url = reverse(
            f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_compare",
            args=(queryset[0].pk,),
        )
        url += f"?compare_to={queryset[1].pk}"

        return redirect(url)

    def delete_view(self, request, object_id, extra_context=None):
        """Do not allow deleting single version objects. Use discard instead."""
        raise PermissionDenied

    @admin.action(
        permissions=["delete"],
        description=_("Delete selected %(verbose_name_plural)s"),
    )
    def delete_selected(self, request, queryset):
        """
        Redirects to a delete versions view based on a users choice
        """
        # Do not allow deleting single version objects. Use discard instead.
        forbidden = queryset.filter(state__in=(PUBLISHED, DRAFT))
        if forbidden.exists():
            self.message_user(
                request,
                _("Draft or published versions cannot be deleted. First unpublish or use discard for drafts."),
                messages.ERROR
            )
            return None

        if request.POST.get("post"):
            # When the user confirms, delete the content objects
            queryset = self.get_content_queryset(queryset)
        return delete_selected(self, request, queryset)

    def get_deleted_objects(self, objs, request):
        """Return the content objects to be deleted"""
        if issubclass(objs.model, Version):
            objs = self.get_content_queryset(objs)
        return super().get_deleted_objects(objs, request)

    def get_content_queryset(self, queryset):
        return self.model._source_model._base_manager.filter(
            pk__in=queryset.values_list("object_id", flat=True)
        )

    def grouper_form_view(self, request):
        """Displays an intermediary page to select a grouper object
        to show versions of.
        """
        language = get_language_from_request(request)
        versionable = versionables.for_content(self.model._source_model)
        context = dict(
            self.admin_site.each_context(request),
            opts=self.model._meta,
            form=grouper_form_factory(self.model._source_model, language, self.admin_site)(),
            title=_("Select {} to view its versions").format(versionable.grouper_model._meta.verbose_name),
        )
        return render(request, "djangocms_versioning/admin/grouper_form.html", context)

    def archive_view(self, request, object_id):
        """Archives the specified version and redirects back to the
        version changelist
        """

        # Check version exists
        version = self.get_object(request, unquote(object_id))
        if version is None:
            return self._get_obj_does_not_exist_redirect(
                request, self.model._meta, object_id
            )

        if not version.can_be_archived():
            self.message_user(request, _("Version cannot be archived"), messages.ERROR)
            return redirect(version_list_url(version.content))
        try:
            version.check_archive(request.user)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return redirect(version_list_url(version.content))

        if request.method != "POST":
            context = {
                "object_name": version.content,
                "version_number": version.number,
                "object_id": object_id,
                "archive_url": reverse(
                    f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_archive",
                    args=(version.content.pk,),
                ),
                "back_url": self.back_link(request, version),
            }
            return render(
                request, "djangocms_versioning/admin/archive_confirmation.html", context
            )
        else:
            # Archive the version
            version.archive(request.user)
            # Display message
            self.message_user(request, _("Version archived"))
        # Redirect
        return redirect(version_list_url(version.content))

    def publish_view(self, request, object_id):
        """Publishes the specified version and redirects back to the
        version changelist
        """
        # This view always changes data so only POST requests should work
        if request.method != "POST":
            return HttpResponseNotAllowed(
                ["POST"], _("This view only supports POST method.")
            )

        # Check version exists
        version = self.get_object(request, unquote(object_id))
        if version is None:
            return self._get_obj_does_not_exist_redirect(
                request, self.model._meta, object_id
            )

        requested_redirect = request.GET.get("next", None)
        if conf.ON_PUBLISH_REDIRECT in ("preview", "published"):
            redirect_url=get_preview_url(version.content)
        else:
            redirect_url=version_list_url(version.content)

        if not version.can_be_published():
            self.message_user(request, _("Version cannot be published"), messages.ERROR)
            return self._internal_redirect(requested_redirect, redirect_url)
        try:
            version.check_publish(request.user)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return self._internal_redirect(requested_redirect, redirect_url)

        # Publish the version
        version.publish(request.user)

        # Display message
        self.message_user(request, _("Version published"))

        # Redirect to published?
        if conf.ON_PUBLISH_REDIRECT == "published":
            if hasattr(version.content, "get_absolute_url"):
                requested_redirect = requested_redirect or version.content.get_absolute_url()

        return self._internal_redirect(requested_redirect, redirect_url)


    def _internal_redirect(self, url, fallback):
        """Helper function to check if the give URL is resolvable
        If resolvable, return the URL; otherwise, returns the fallback URL.
        """
        if not url:
            return redirect(fallback)

        try:
            resolve(url)
        except Resolver404:
            return redirect(fallback)

        return redirect(url)


    def unpublish_view(self, request, object_id):
        """Unpublishes the specified version and redirects back to the
        version changelist
        """
        # Check version exists
        version = self.get_object(request, unquote(object_id))
        if version is None:
            return self._get_obj_does_not_exist_redirect(
                request, self.model._meta, object_id
            )

        if conf.ON_PUBLISH_REDIRECT in ("preview", "published"):
            redirect_url=get_preview_url(version.content)
        else:
            redirect_url=version_list_url(version.content)

        if not version.can_be_unpublished():
            self.message_user(
                request, _("Version cannot be unpublished"), messages.ERROR
            )
            return redirect(redirect_url)
        try:
            version.check_unpublish(request.user)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return redirect(redirect_url)

        if request.method != "POST":
            context = {
                "object_name": version.content,
                "version_number": version.number,
                "object_id": object_id,
                "unpublish_url": reverse(
                    f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_unpublish",
                    args=(version.content.pk,),
                ),
                "back_url": self.back_link(request, version),
            }
            extra_context = OrderedDict(
                [
                    (key, func(request, version))
                    for key, func in _cms_extension()
                    .add_to_context.get("unpublish", {})
                    .items()
                ]
            )
            context.update({"extra_context": extra_context})
            return render(
                request,
                "djangocms_versioning/admin/unpublish_confirmation.html",
                context,
            )
        else:
            # Unpublish the version
            version.unpublish(request.user)
            # Display message
            self.message_user(request, _("Version unpublished"))
        # Redirect
        return redirect(redirect_url)

    def _get_edit_redirect_version(self, request, version):
        """Helper method to get the latest draft or create one if one does not exist."""
        # If published then there's extra things to do...
        if version.state == PUBLISHED:
            # First check there is no draft record for this grouper
            # already.
            pks_for_grouper = version.versionable.for_content_grouping_values(
                version.content
            ).values_list("pk", flat=True)
            content_type = ContentType.objects.get_for_model(version.content)
            drafts = Version.objects.filter(
                object_id__in=pks_for_grouper, content_type=content_type, state=DRAFT
            )
            if drafts.exists():
                # There is a draft record so people should be editing
                # the draft record not the published one. Redirect to draft.
                draft = drafts.first()
                # Run edit checks for the found draft as well
                draft.check_edit_redirect(request.user)
                if conf.LOCK_VERSIONS:
                    create_version_lock(version, request.user)
                return draft
            # If there is no draft record then create a new version
            # that's a draft with the content copied over
            return version.copy(request.user)
        elif version.state == DRAFT:
            if conf.LOCK_VERSIONS:
                create_version_lock(version, request.user)
            # Return current version as it is a draft
            return version

    def edit_redirect_view(self, request, object_id):
        """Redirects to the admin change view and creates a draft version
        if no draft exists yet.
        """
        # This view always changes data so only POST requests should work
        if request.method != "POST":
            return HttpResponseNotAllowed(
                ["POST"], _("This view only supports POST method.")
            )

        version = self.get_object(request, unquote(object_id))
        if version is None:
            raise Http404

        try:
            version.check_edit_redirect(request.user)
            target = self._get_edit_redirect_version(request, version)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return redirect(version_list_url(version.content))

        # Redirect
        return redirect(get_editable_url(target.content, request.GET.get("force_admin")))

    def revert_view(self, request, object_id):
        """Reverts to the specified version i.e. creates a draft from it."""
        version = self.get_object(request, unquote(object_id))

        if version is None:
            raise Http404

        try:
            version.check_revert(request.user)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return redirect(version_list_url(version.content))

        pks_for_grouper = version.versionable.for_content_grouping_values(
            version.content
        ).values_list("pk", flat=True)
        drafts = Version.objects.filter(
            object_id__in=pks_for_grouper,
            content_type=version.content_type,
            state=DRAFT,
        )

        draft_version = None
        if drafts.exists():
            draft_version = drafts.first()

        if request.method != "POST":
            context = {
                "object_name": version.content,
                "version_number": version.number,
                "draft_version": draft_version,
                "object_id": object_id,
                "revert_url": reverse(
                    f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_revert",
                    args=(version.content.pk,),
                ),
                "back_url": self.back_link(request, version),
            }
            return render(
                request, "djangocms_versioning/admin/revert_confirmation.html", context
            )
        else:

            if draft_version and request.POST.get("archive"):
                draft_version.archive(request.user)

            if draft_version and request.POST.get("discard"):
                draft_version.delete()

            version = version.copy(request.user)
            # Redirect
            return redirect(version_list_url(version.content))

    def discard_view(self, request, object_id):
        """Discards the specified version"""
        version = self.get_object(request, unquote(object_id))
        if version is None:
            raise Http404

        try:
            version.check_discard(request.user)
        except ConditionFailed as e:
            self.message_user(request, force_str(e), messages.ERROR)
            return redirect(version_list_url(version.content))

        if request.method != "POST":
            context = {
                "object_name": version.content,
                "version_number": version.number,
                "draft_version": version,
                "object_id": object_id,
                "revert_url": reverse(
                    f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_revert",
                    args=(version.content.pk,),
                ),
                "back_url": self.back_link(request, version),
            }
            return render(
                request, "djangocms_versioning/admin/discard_confirmation.html", context
            )

        version_url = version_list_url(version.content)
        if request.POST.get("discard"):
            ModelClass = version.content.__class__
            deleted = version.delete()
            if deleted[1]["last"]:
                version_url = get_admin_url(ModelClass, "changelist")
                self.message_user(request, _("The last version has been deleted"))

        return redirect(version_url)

    def compare_view(self, request, object_id):
        """Compares two versions
        """
        # Get version 1 (the version we're comparing against)
        v1 = self.get_object(request, unquote(object_id))
        if v1 is None:
            return self._get_obj_does_not_exist_redirect(
                request, self.model._meta, object_id
            )
        persist_params = {
            get_cms_setting("CMS_TOOLBAR_URL__DISABLE"): 1,
            get_cms_setting("CMS_TOOLBAR_URL__PERSIST"): 0,
        }
        v1_preview_url = get_preview_url(v1.content)
        v1_preview_url = add_url_parameters(v1_preview_url, **persist_params)
        # Get the list of versions for the grouper. This is for use
        # in the dropdown to choose a version.
        version_list = Version.objects.filter_by_content_grouping_values(
            v1.content
        ).order_by("-number")
        # Add the above to context
        context = {
            "version_list": version_list,
            "v1": v1,
            "v1_preview_url": v1_preview_url,
            "return_url": self.back_link(request, v1),
        }

        # Now check if version 2 has been specified and add to context
        # if yes
        if "compare_to" in request.GET:
            v2 = self.get_object(request, unquote(request.GET["compare_to"]))
            if v2 is None:
                return self._get_obj_does_not_exist_redirect(
                    request, self.model._meta, request.GET["compare_to"]
                )
            else:
                v2_preview_url = get_preview_url(v2.content)
                context.update(
                    {
                        "v2": v2,
                        "v2_preview_url": add_url_parameters(v2_preview_url, **persist_params),
                    }
                )
        return TemplateResponse(
            request, "djangocms_versioning/admin/compare.html", context
        )

    def unlock_view(self, request, object_id):
        """
        Unlock a locked version
        """
        # Only active if LOCK_VERISONS is set
        if not conf.LOCK_VERSIONS:
            raise Http404()

        # This view always changes data so only POST requests should work
        if request.method != "POST":
            return HttpResponseNotAllowed(["POST"], _("This view only supports POST method."))

        # Check version exists
        version = self.get_object(request, unquote(object_id))
        if version is None:
            return self._get_obj_does_not_exist_redirect(
                request, self.model._meta, object_id)

        # Raise 404 if not locked
        if version.state != DRAFT:
            raise Http404

        # Check that the user has unlock permission
        if not request.user.has_perm("djangocms_versioning.delete_versionlock"):
            return HttpResponseForbidden(force_str(_("You do not have permission to remove the version lock")))

        # Unlock the version
        remove_version_lock(version)
        # Display message
        messages.success(request, _("Version unlocked"))

        # Send an email notification
        notify_version_author_version_unlocked(version, request.user)

        # Redirect
        url = version_list_url(version.content)
        return redirect(url)

    @staticmethod
    def back_link(request, version=None):
        back_url = request.GET.get("back", None)
        if back_url:
            try:
                # Is return url a valid url?
                resolve(urlparse(back_url)[2])
            except Resolver404:
                # If not ignore
                back_url = None
        return back_url or (version_list_url(version.content) if version else None)

    def changelist_view(self, request, extra_context=None):
        """Handle grouper filtering on the changelist"""
        if not request.GET:
            # redirect to grouper form when there's no GET parameters
            opts = self.model._meta
            return redirect(
                reverse(f"admin:{opts.app_label}_{opts.model_name}_grouper")
            )
        extra_context = extra_context or {}
        versionable = versionables.for_content(self.model._source_model)

        try:
            grouper = versionable.get_grouper_with_fallbacks(
                int(request.GET.get(versionable.grouper_field_name))
            )
        except (TypeError, ValueError):
            grouper = None
        else:
            if grouper is None:
                # no exception and no grouper, thus the querydict is invalid
                raise Http404

        if grouper:
            # CAVEAT: as the breadcrumb trails expect a value for latest content in the template
            extra_context["latest_content"] = ({"pk": None})

            extra_context.update(
                grouper=grouper,
                title=_('Displaying versions of "{grouper}"').format(grouper=grouper),
            )
            breadcrumb_opts = self.model._source_model._meta
            extra_context["breadcrumb_opts"] = breadcrumb_opts
            # Check if custom breadcrumb template defined, otherwise
            # fallback on default
            breadcrumb_templates = [
                f"admin/djangocms_versioning/{breadcrumb_opts.app_label}/{breadcrumb_opts.model_name}/versioning_breadcrumbs.html",
                "admin/djangocms_versioning/versioning_breadcrumbs.html",
            ]
            extra_context["breadcrumb_template"] = select_template(breadcrumb_templates)

        response = super().changelist_view(request, extra_context)

        # This is a slightly hacky way of accessing the instance of
        # the changelist that the admin changelist_view instantiates.
        # We do this to make sure that the latest content object is
        # picked from the same queryset as is being displayed in the
        # version table.
        if grouper and response.status_code == 200:
            # Catch the edge case where a grouper can have empty contents
            # when additional filters are present and the result set will be
            # empty for the additional values.
            try:
                response.context_data["latest_content"] = (
                    response.context_data["cl"].get_queryset(request)
                        .latest("created")
                        .content
                )
            except (ObjectDoesNotExist, KeyError):
                pass
        return response

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return [
            path(
                "select/",
                self.admin_site.admin_view(self.grouper_form_view),
                name="{}_{}_grouper".format(*info),
            ),
            path(
                "<path:object_id>/archive/",
                self.admin_site.admin_view(self.archive_view),
                name="{}_{}_archive".format(*info),
            ),
            path(
                r"<path:object_id>/publish/",
                self.admin_site.admin_view(self.publish_view),
                name="{}_{}_publish".format(*info),
            ),
            path(
                "<path:object_id>/unpublish/",
                self.admin_site.admin_view(self.unpublish_view),
                name="{}_{}_unpublish".format(*info),
            ),
            path(
                "<path:object_id>/edit-redirect/",
                self.admin_site.admin_view(self.edit_redirect_view),
                name="{}_{}_edit_redirect".format(*info),
            ),
            path(
                "<path:object_id>/revert/",
                self.admin_site.admin_view(self.revert_view),
                name="{}_{}_revert".format(*info),
            ),
            path(
                "<path:object_id>/compare/",
                self.admin_site.admin_view(self.compare_view),
                name="{}_{}_compare".format(*info),
            ),
            path(
                "<path:object_id>/discard/",
                self.admin_site.admin_view(self.discard_view),
                name="{}_{}_discard".format(*info),
            ),
            path(
                "<path:object_id>/unlock/",
                self.admin_site.admin_view(self.unlock_view),
                name="{}_{}_unlock".format(*info),
            ),
        ] + super().get_urls()

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        """Disable change view access"""
        if obj is not None:
            return False
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        if obj is None:
            return conf.ALLOW_DELETING_VERSIONS and super().has_delete_permission(request, obj)
        content_admin = self.admin_site._registry[self.model._source_model]
        return all((
            conf.ALLOW_DELETING_VERSIONS,
            super().has_delete_permission(request, obj),
            content_admin.has_delete_permission(request, obj.content),
        ))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-07-24 12:42
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("djangocms_versioning", "0001_initial")]

    operations = [migrations.DeleteModel(name="Campaign")]



from collections import OrderedDict

from cms.app_base import CMSAppConfig
from cms.utils.i18n import get_language_tuple

from djangocms_versioning.datastructures import VersionableItem, default_copy

from .models import PollContent


def unpublish_warning(request, version, *args, **kwargs):
    return "<b>Warning: Polls don't like to be unpublished!</b>"


def poll_modifier(obj, field):
    return "[TEST]" + getattr(obj, field)


poll_modifier.short_description = "Title"


class PollsCMSConfig(CMSAppConfig):
    djangocms_versioning_enabled = True
    versioning = [
        VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            extra_grouping_fields=["language"],
            version_list_filter_lookups={"language": get_language_tuple},
            copy_function=default_copy,
            preview_url=PollContent.get_preview_url,
        )
    ]
    versioning_add_to_confirmation_context = {
        "unpublish": OrderedDict([("warning", unpublish_warning)])
    }
    extended_admin_field_modifiers = [
        {PollContent: {"text": poll_modifier}},
    ]



from cms.test_utils.testcases import CMSTestCase
from django.conf import settings
from django.db import models
from django.test import override_settings

from djangocms_versioning import constants, models as versioning_models
from djangocms_versioning.test_utils import factories


class DeletionTestCase(CMSTestCase):
    @override_settings(DJANGOCMS_VERSIONING_ALLOW_DELETING_VERSIONS=False)
    def test_deletion_not_possible(self):
        # Since djangocms_versionings.models stores the setting, we need to update that copy
        versioning_models.ALLOW_DELETING_VERSIONS = settings.DJANGOCMS_VERSIONING_ALLOW_DELETING_VERSIONS
        poll = factories.PollFactory()
        version1 = factories.PollVersionFactory(
            content__poll=poll,
            content__language="en",
        )
        pk1 = version1.pk
        # Now publish and then edit redirect to create a draft on top of published version
        version1.publish(user=self.get_superuser())
        self.assertEqual(versioning_models.Version.objects.get(pk=pk1).state, constants.PUBLISHED)

        version2 = version1.copy(created_by=self.get_superuser())
        version2.save()

        # Check of source field is set
        self.assertIsNotNone(version2.source)

        # try deleting and see if error is raised
        self.assertRaises(models.deletion.ProtectedError,
                          versioning_models.Version.objects.get(pk=pk1).content.delete)

    @override_settings(DJANGOCMS_VERSIONING_ALLOW_DELETING_VERSIONS=True)
    def test_deletion_possible(self):
        # Since djangocms_versionings.models stores the setting, we need to update that copy
        versioning_models.ALLOW_DELETING_VERSIONS = settings.DJANGOCMS_VERSIONING_ALLOW_DELETING_VERSIONS
        poll = factories.PollFactory()
        version1 = factories.PollVersionFactory(
            content__poll=poll,
            content__language="en",
        )
        pk1 = version1.pk
        # Now publish and then edit redirect to create a draft on top of published version
        version1.publish(user=self.get_superuser())
        self.assertEqual(versioning_models.Version.objects.get(pk=pk1).state, constants.PUBLISHED)

        version2 = version1.copy(created_by=self.get_superuser())
        version2.save()

        # Check of source field is set
        self.assertIsNotNone(version2.source)

        # try deleting and see if error is raised
        versioning_models.Version.objects.get(pk=pk1).content.delete()
        self.assertIsNone(versioning_models.Version.objects.get(pk=version2.pk).source)
